class NoTokenLimitFound(Exception):
    pass


def get_default_token_limit(model_id: str) -> int:
    """
    Get token limits for known models.
    """

    # TODO: Add ability to register more default models, e.g. in Django settings.

    match model_id:
        case "gpt-3.5-turbo":
            return 4096
        case "gpt-3.5-turbo-16k":
            return 16385
        case "gpt-4":
            return 8192
        case "gpt-4-32k":
            return 32768
        case _:
            raise NoTokenLimitFound(model_id)



from abc import ABCMeta
from dataclasses import dataclass
from enum import Enum
from typing import (
    Any,
    ClassVar,
    Generic,
    NotRequired,
    Protocol,
    Required,
    Self,
    TypedDict,
    TypeVar,
)

from django.core.exceptions import ImproperlyConfigured
from django.core.files import File

from .. import tokens
from ..types import (
    AIResponse,
    TextSplitterLengthCalculatorProtocol,
    TextSplitterProtocol,
)


class BackendFeature(Enum):
    TEXT_COMPLETION = "TEXT_COMPLETION"
    IMAGE_DESCRIPTION = "IMAGE_DESCRIPTION"


class BaseAIBackendConfigSettings(TypedDict):
    MODEL_ID: Required[str]
    TOKEN_LIMIT: NotRequired[int | None]


AIBackendConfigSettings = TypeVar(
    "AIBackendConfigSettings", bound=BaseAIBackendConfigSettings, contravariant=True
)


class ConfigClassProtocol(Protocol[AIBackendConfigSettings]):
    @classmethod
    def from_settings(cls, config: AIBackendConfigSettings, **kwargs: Any) -> Self:
        ...


@dataclass(kw_only=True)
class BaseAIBackendConfig(ConfigClassProtocol[AIBackendConfigSettings]):
    model_id: str
    token_limit: int
    text_splitter_class: type[TextSplitterProtocol]
    text_splitter_length_calculator_class: type[TextSplitterLengthCalculatorProtocol]

    @classmethod
    def from_settings(
        cls,
        config: AIBackendConfigSettings,
        *,
        text_splitter_class: type[TextSplitterProtocol],
        text_splitter_length_calculator_class: type[
            TextSplitterLengthCalculatorProtocol
        ],
        **kwargs: Any,
    ) -> Self:
        token_limit = cls.get_token_limit(
            model_id=config["MODEL_ID"], custom_value=config.get("TOKEN_LIMIT")
        )

        return cls(
            model_id=config["MODEL_ID"],
            token_limit=token_limit,
            text_splitter_class=text_splitter_class,
            text_splitter_length_calculator_class=text_splitter_length_calculator_class,
            **kwargs,
        )

    @classmethod
    def get_token_limit(cls, *, model_id: str, custom_value: int | None) -> int:
        if custom_value is not None:
            try:
                return int(custom_value)
            except ValueError as e:
                raise ImproperlyConfigured(
                    f'"TOKEN_LIMIT" is not an "int", it is a "{type(custom_value)}".'
                ) from e
        try:
            return tokens.get_default_token_limit(model_id=model_id)
        except tokens.NoTokenLimitFound as e:
            raise ImproperlyConfigured(
                f'"TOKEN_LIMIT" is not configured for model "{model_id}".'
            ) from e


AIBackendConfig = TypeVar("AIBackendConfig", bound=BaseAIBackendConfig)


class AIBackend(Generic[AIBackendConfig], metaclass=ABCMeta):
    config_cls: ClassVar[type[ConfigClassProtocol]]
    config: AIBackendConfig

    def __init__(
        self,
        *,
        config: AIBackendConfig,
    ) -> None:
        self.config = config

    def prompt_with_context(
        self, *, pre_prompt: str, context: str, post_prompt: str | None = None
    ) -> AIResponse:
        """
        Given a prompt and a context, return a response.
        """
        raise NotImplementedError("This backend does not support text completion")

    def get_text_splitter(self) -> TextSplitterProtocol:
        return self.config.text_splitter_class(
            chunk_size=self.config.token_limit,
            length_function=self.get_splitter_length_calculator().get_splitter_length,
        )

    def get_splitter_length_calculator(self) -> TextSplitterLengthCalculatorProtocol:
        return self.config.text_splitter_length_calculator_class()

    def describe_image(self, *, image_file: File, prompt: str) -> AIResponse:
        raise NotImplementedError("This backend does not support image description")



from collections.abc import Sequence
from typing import NotRequired, Required, TypedDict


class PromptDict(TypedDict):
    default_prompt_id: Required[int]
    label: Required[str]
    description: NotRequired[str]
    prompt: Required[str]
    method: Required[str]


DEFAULT_PROMPTS: Sequence[PromptDict] = [
    {
        "default_prompt_id": 1,  # A unique ID used to identify and manage default prompts
        "label": "AI Correction",
        "description": "Correct grammar and spelling",
        "prompt": (
            "You are assisting a user in writing content for their website. "
            "The user has provided some text (following the colon). "
            "Return the provided text but with corrected grammar, spelling and punctuation. "
            "Do not add additional punctuation, quotation marks or change any words:"
        ),
        "method": "replace",
    },
    {
        "default_prompt_id": 2,  # A unique ID used to identify and manage default prompts
        "label": "AI Completion",
        "description": "Get help writing more content based on what you've written",
        "prompt": (
            "You are assisting a user in writing content for their website. "
            "The user has provided some initial text (following the colon). "
            "Assist the user in writing the remaining content:"
        ),
        "method": "append",
    },
]



import logging
from collections.abc import Callable

from ..types import TextSplitterLengthCalculatorProtocol, TextSplitterProtocol

logger = logging.getLogger(__name__)


class DummyTextSplitter(TextSplitterProtocol):
    def __init__(
        self, *, chunk_size: int, length_function: Callable[[str], int]
    ) -> None:
        pass

    def split_text(self, text: str) -> list[str]:
        # Don't do any splitting.
        return [text]


class DummyLengthCalculator(TextSplitterLengthCalculatorProtocol):
    def get_splitter_length(self, text: str) -> int:
        return len(text)



from functools import cached_property

from django import forms
from django.core.exceptions import ValidationError
from django.utils.translation import gettext_lazy as _
from wagtail.admin.staticfiles import versioned_static
from wagtail.images.forms import BaseImageForm


class PromptTextField(forms.CharField):
    default_error_messages = {
        "required": _(
            "No text provided - please enter some text before using AI features."
        ),
    }


class PromptUUIDField(forms.UUIDField):
    default_error_messages = {
        "required": _("Invalid prompt provided."),
        "invalid": _("Invalid prompt provided."),
    }


class ApiForm(forms.Form):
    def errors_for_json_response(self) -> str:
        errors_for_response = []
        for _field, errors in self.errors.get_json_data().items():
            for error in errors:
                errors_for_response.append(error["message"])

        return " \n".join(errors_for_response)


class PromptForm(ApiForm):
    text = PromptTextField()
    prompt = PromptUUIDField()

    def clean_prompt(self):
        prompt_uuid = self.cleaned_data["prompt"]
        if prompt_uuid.version != 4:
            raise ValidationError(
                self.fields["prompt"].error_messages["invalid"], code="invalid"
            )

        return prompt_uuid


class DescribeImageApiForm(ApiForm):
    image_id = forms.CharField()
    maxlength = forms.IntegerField(required=False, min_value=0, max_value=4096)


class DescribeImageForm(BaseImageForm):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.instance and self.instance.pk:
            widget = self.fields["title"].widget
            widget.attrs["data-wagtailai-image-id"] = str(self.instance.pk)
            widget.attrs["data-wagtailai-button-title"] = _("Describe image using AI")
            widget.template_name = "wagtail_ai/widgets/image_title.html"

    @cached_property
    def media(self):
        return super().media + forms.Media(
            js=[
                versioned_static("wagtail_ai/image_description.js"),
            ],
            css={
                "all": [versioned_static("wagtail_ai/image_description.css")],
            },
        )



from django.apps import AppConfig


class WagtailAiAppConfig(AppConfig):
    default_auto_field = "django.db.models.AutoField"
    label = "wagtail_ai"
    name = "wagtail_ai"
    verbose_name = "Wagtail AI"



import warnings
from collections.abc import Mapping
from dataclasses import dataclass
from typing import NotRequired, Required, TypedDict, cast

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.utils.module_loading import import_string

from ..text_splitters.langchain import LangchainRecursiveCharacterTextSplitter
from ..text_splitters.length import NaiveTextSplitterCalculator
from ..types import TextSplitterLengthCalculatorProtocol, TextSplitterProtocol
from ..utils import deprecation
from .base import AIBackend, BackendFeature, BaseAIBackendConfigSettings


class TextSplittingSettingsDict(TypedDict):
    SPLITTER_CLASS: NotRequired[str]
    SPLITTER_LENGTH_CALCULATOR_CLASS: NotRequired[str]


class InvalidAIBackendError(ImproperlyConfigured):
    def __init__(self, alias):
        super().__init__(f"Invalid AI backend: {alias}")


class AIBackendSettingsDict(TypedDict):
    CONFIG: Required[BaseAIBackendConfigSettings]
    CLASS: Required[str]
    TEXT_SPLITTING: NotRequired[TextSplittingSettingsDict]


def get_ai_backends_settings() -> Mapping[str, AIBackendSettingsDict]:
    try:
        return settings.WAGTAIL_AI["BACKENDS"]
    except AttributeError:
        pass
    try:
        legacy_settings = settings.WAGTAIL_AI_BACKENDS
    except AttributeError:
        pass
    else:
        warnings.warn(
            'settings.WAGTAIL_AI_BACKENDS is deprecated. Please use "WAGTAIL_AI = {"BACKENDS": {"default": {...}}}" instead',
            deprecation.WagtailAISettingsDeprecationWarning,
            stacklevel=1,
        )
        return legacy_settings
    return {
        "default": {
            "CLASS": "wagtail_ai.ai.llm.LLMBackend",
            "CONFIG": {
                "MODEL_ID": "gpt-3.5-turbo",
            },
        }
    }


def _validate_backend_settings(*, settings: AIBackendSettingsDict, alias: str) -> None:
    if "CONFIG" not in settings:
        raise ImproperlyConfigured(
            'AI backend settings for "{alias}": Missing "CONFIG".'
        )
    if not isinstance(settings["CONFIG"], Mapping):
        raise ImproperlyConfigured(
            'AI backend settings for "{alias}": "CONFIG" is not a Mapping.'
        )
    if "MODEL_ID" not in settings["CONFIG"]:
        raise ImproperlyConfigured(
            'AI backend settings for "{alias}": "MODEL_ID" is missing in "CONFIG".'
        )


def get_ai_backend_settings(alias: str) -> AIBackendSettingsDict:
    backends_settings = get_ai_backends_settings()
    try:
        return backends_settings[alias]
    except (KeyError, ImportError) as e:
        raise InvalidAIBackendError(alias) from e

    _validate_backend_settings(settings=backends_settings, alias=alias)


def _get_default_text_splitter_class() -> type[LangchainRecursiveCharacterTextSplitter]:
    return LangchainRecursiveCharacterTextSplitter


def _get_default_text_splitter_length_class() -> type[NaiveTextSplitterCalculator]:
    return NaiveTextSplitterCalculator


@dataclass(kw_only=True)
class _TextSplitterConfig:
    splitter_class: type[TextSplitterProtocol]
    splitter_length_calculator_class: type[TextSplitterLengthCalculatorProtocol]


def _get_text_splitter_config(
    *, backend_alias: str, config: TextSplittingSettingsDict
) -> _TextSplitterConfig:
    splitter_class_path = config.get("SPLITTER_CLASS")
    length_calculator_class_path = config.get("SPLITTER_LENGTH_CALCULATOR_CLASS")

    # Text splitter - class that splits text into chunks of a given size.
    if splitter_class_path is not None:
        try:
            splitter_class = cast(
                type[TextSplitterProtocol], import_string(splitter_class_path)
            )
        except ImportError as e:
            raise ImproperlyConfigured(
                f'Cannot import "SPLITTER_CLASS" ("{splitter_class_path}") for backend "{backend_alias}".'
            ) from e
    else:
        splitter_class = _get_default_text_splitter_class()

    # Text splitter length calculator - class that calculates the number of token in a given text.
    if length_calculator_class_path is not None:
        try:
            length_calculator_class = cast(
                type[TextSplitterLengthCalculatorProtocol],
                import_string(length_calculator_class_path),
            )
        except ImportError as e:
            raise ImproperlyConfigured(
                f'Cannot import "SPLITTER_LENGTH_CALCULATOR_CLASS" ("{length_calculator_class_path}") for backend "{backend_alias}".'
            ) from e
    else:
        length_calculator_class = _get_default_text_splitter_length_class()

    return _TextSplitterConfig(
        splitter_class=splitter_class,
        splitter_length_calculator_class=length_calculator_class,
    )


def get_ai_backend(alias: str) -> AIBackend:
    backend_dict = get_ai_backend_settings(alias)

    if "CLASS" not in backend_dict:
        raise ImproperlyConfigured(
            f'"AI backend "{alias}" settings: "CLASS" is missing in the configuration.'
        )

    try:
        ai_backend_cls = cast(type[AIBackend], import_string(backend_dict["CLASS"]))
    except ImportError as e:
        raise InvalidAIBackendError(
            f'"AI backend "{alias}" settings: "CLASS" ("{backend_dict["CLASS"]}") is not importable.'
        ) from e

    backend_settings = backend_dict["CONFIG"]
    text_splitting = _get_text_splitter_config(
        backend_alias=alias,
        config=backend_dict.get("TEXT_SPLITTING", {}),
    )
    config = ai_backend_cls.config_cls.from_settings(
        backend_settings,
        text_splitter_class=text_splitting.splitter_class,
        text_splitter_length_calculator_class=text_splitting.splitter_length_calculator_class,
    )

    return ai_backend_cls(config=config)


class BackendNotFound(Exception):
    pass


def get_backend(feature: BackendFeature = BackendFeature.TEXT_COMPLETION) -> AIBackend:
    match feature:
        case BackendFeature.TEXT_COMPLETION:
            alias = settings.WAGTAIL_AI.get("TEXT_COMPLETION_BACKEND", "default")
        case BackendFeature.IMAGE_DESCRIPTION:
            alias = settings.WAGTAIL_AI.get("IMAGE_DESCRIPTION_BACKEND")
        case _:
            alias = None

    if alias is None:
        raise BackendNotFound(f"No backend found for {feature.name}")

    return get_ai_backend(alias)



import random
import time
from collections.abc import Generator, Iterator
from dataclasses import dataclass
from typing import Any, NotRequired, Self

from django.core.exceptions import ImproperlyConfigured
from django.core.files import File

from .base import (
    AIBackend,
    AIResponse,
    BaseAIBackendConfig,
    BaseAIBackendConfigSettings,
)


class EchoResponse(AIResponse):
    _text: str | None = None
    response_iterator: Iterator[str]

    def __init__(self, response_iterator: Iterator[str]) -> None:
        self.response_iterator = response_iterator

    def __iter__(self) -> Iterator[str]:
        return self.response_iterator

    def text(self) -> str:
        if self._text is not None:
            return self._text
        self._text = " ".join(self.response_iterator)
        return self._text


@dataclass(kw_only=True)
class EchoBackendSettingsDict(BaseAIBackendConfigSettings):
    MAX_WORD_SLEEP_SECONDS: NotRequired[int]


@dataclass(kw_only=True)
class EchoBackendConfig(BaseAIBackendConfig[EchoBackendSettingsDict]):
    max_word_sleep_seconds: int

    @classmethod
    def from_settings(cls, config: EchoBackendSettingsDict, **kwargs: Any) -> Self:
        max_word_sleep_seconds = config.get("MAX_WORD_SLEEP_SECONDS")
        if max_word_sleep_seconds is None:
            max_word_sleep_seconds = 0
        try:
            max_word_sleep_seconds = int(max_word_sleep_seconds)
        except ValueError as e:
            raise ImproperlyConfigured(
                f'"MAX_WORD_SLEEP_SECONDS" is not an "int", it is a "{type(max_word_sleep_seconds)}".'
            ) from e
        kwargs.setdefault("max_word_sleep_seconds", max_word_sleep_seconds)

        return super().from_settings(config, **kwargs)


class EchoBackend(AIBackend[EchoBackendConfig]):
    config_cls = EchoBackendConfig

    def prompt_with_context(
        self, *, pre_prompt: str, context: str, post_prompt: str | None = None
    ) -> AIResponse:
        return self.get_response(
            ["This", "is", "an", "echo", "backend:", *context.split()]
        )

    def describe_image(self, *, image_file: File, prompt: str) -> AIResponse:
        return self.get_response(
            ["This", "is", "an", "echo", "backend:", image_file.name]
        )

    def get_response(self, words):
        def response_iterator() -> Generator[str, None, None]:
            for word in words:
                if (
                    self.config.max_word_sleep_seconds is not None
                    and self.config.max_word_sleep_seconds > 0
                ):
                    time.sleep(
                        random.random()
                        * random.randint(0, self.config.max_word_sleep_seconds)
                    )
                yield word

        return EchoResponse(response_iterator())



import base64
import os
from collections.abc import Iterator
from dataclasses import dataclass
from typing import Any, NotRequired, Self

import requests
from django.core.files import File

from wagtail_ai.types import AIResponse

from .base import AIBackend, BaseAIBackendConfig, BaseAIBackendConfigSettings


class OpenAIBackendConfigSettingsDict(BaseAIBackendConfigSettings):
    TIMEOUT_SECONDS: NotRequired[int | None]
    OPENAI_API_KEY: NotRequired[str | None]


@dataclass(kw_only=True)
class OpenAIBackendConfig(BaseAIBackendConfig[OpenAIBackendConfigSettingsDict]):
    timeout_seconds: int
    openai_api_key: str | None

    @classmethod
    def from_settings(
        cls, config: OpenAIBackendConfigSettingsDict, **kwargs: Any
    ) -> Self:
        timeout_seconds = config.get("TIMEOUT_SECONDS")
        if timeout_seconds is None:
            timeout_seconds = 15
        kwargs.setdefault("timeout_seconds", timeout_seconds)

        kwargs.setdefault("openai_api_key", config.get("OPENAI_API_KEY"))

        return super().from_settings(config, **kwargs)


class OpenAIResponse(AIResponse):
    def __init__(self, response: requests.Response):
        self.response = response
        self._text = response.json()["choices"][0]["message"]["content"]

    def __iter__(self) -> Iterator[str]:
        yield self._text

    def text(self) -> str:
        return self._text

    def __str__(self):
        return self.text()


class OpenAIBackend(AIBackend[OpenAIBackendConfig]):
    config_cls = OpenAIBackendConfig

    def prompt_with_context(
        self, *, pre_prompt: str, context: str, post_prompt: str | None = None
    ) -> OpenAIResponse:
        messages = [
            {"role": "system", "content": [{"type": "text", "text": pre_prompt}]},
            {"role": "user", "content": [{"type": "text", "text": context}]},
        ]

        if post_prompt is not None:
            messages.append(
                {"role": "system", "content": [{"type": "text", "text": post_prompt}]}
            )

        return self.chat_completions(messages)

    def describe_image(self, *, image_file: File, prompt: str) -> OpenAIResponse:
        if not prompt:
            raise ValueError("Prompt must not be empty.")
        with image_file.open() as f:
            base64_image = base64.b64encode(f.read()).decode("utf-8")

        return self.chat_completions(
            messages=[
                {
                    "role": "user",
                    "content": [
                        {
                            "type": "text",
                            "text": prompt,
                        },
                        {
                            "type": "image_url",
                            "image_url": {
                                "url": f"data:image/jpeg;base64,{base64_image}"
                            },
                        },
                    ],
                },
            ],
        )

    def chat_completions(self, messages: list[dict[str, Any]]) -> OpenAIResponse:
        headers = {
            "Content-Type": "application/json",
            "Authorization": f"Bearer {self.get_openai_api_key()}",
        }
        payload = {
            "model": self.config.model_id,
            "messages": messages,
            "max_tokens": self.config.token_limit,
        }
        response = requests.post(
            "https://api.openai.com/v1/chat/completions",
            headers=headers,
            json=payload,
            timeout=self.config.timeout_seconds,
        )

        response.raise_for_status()
        return OpenAIResponse(response)

    def get_openai_api_key(self) -> str:
        if config_key := self.config.openai_api_key:
            return config_key

        if env_key := os.environ.get("OPENAI_API_KEY"):
            return env_key

        raise RuntimeError(
            "Cannot find OpenAI API key. Please set the OPENAI_API_KEY environment"
            " variable or the OPENAI_API_KEY backend config value in Django settings."
        )



import uuid
from typing import NotRequired, Required, TypedDict

from django.urls import include, path, reverse
from django.utils.html import json_script
from django.utils.safestring import mark_safe
from django.views.i18n import JavaScriptCatalog
from wagtail import hooks
from wagtail.admin.rich_text.editors.draftail.features import ControlFeature

from .models import Prompt
from .views import describe_image, prompt_viewset, text_completion


@hooks.register("register_admin_urls")  # type: ignore
def register_admin_urls():
    urls = [
        path(
            "jsi18n/",
            JavaScriptCatalog.as_view(packages=["wagtail_ai"]),
            name="javascript_catalog",
        ),
        path(
            "text_completion/",
            text_completion,
            name="text_completion",
        ),
        path(
            "describe_image/",
            describe_image,
            name="describe_image",
        ),
    ]

    return [
        path(
            "ai/",
            include(
                (urls, "wagtail_ai"),
                namespace="wagtail_ai",
            ),
        )
    ]


@hooks.register("register_rich_text_features")  # type: ignore
def register_ai_feature(features):
    feature_name = "ai"
    features.default_features.append(feature_name)

    features.register_editor_plugin(
        "draftail",
        feature_name,
        ControlFeature(
            {
                "type": feature_name,
            },
            js=["wagtail_ai/draftail.js"],
            css={"all": ["wagtail_ai/draftail.css"]},
        ),
    )


class PromptDict(TypedDict):
    # Fields should match the Prompt type defined in custom.d.ts
    # be careful not to expose any sensitive or exploitable data here.
    uuid: Required[uuid.UUID]
    label: Required[str]
    description: NotRequired[str]
    prompt: Required[str]
    method: Required[str]


def _serialize_prompt(prompt: Prompt) -> PromptDict:
    return {
        "uuid": prompt.uuid,
        "label": prompt.label,
        "description": prompt.description,
        "prompt": prompt.prompt_value,
        "method": prompt.method,
    }


def get_prompts():
    return [_serialize_prompt(prompt) for prompt in Prompt.objects.all()]


@hooks.register("insert_global_admin_js")  # type: ignore
def ai_editor_js():
    config = {
        "aiPrompts": get_prompts(),
        "urls": {
            "TEXT_COMPLETION": reverse("wagtail_ai:text_completion"),
            "DESCRIBE_IMAGE": reverse("wagtail_ai:describe_image"),
        },
    }

    return mark_safe(json_script(config, "wagtail-ai-config"))


@hooks.register("register_admin_viewset")  # type: ignore
def register_viewset():
    return prompt_viewset



import decimal
import logging
import math
import re

from ..types import TextSplitterLengthCalculatorProtocol

logger = logging.getLogger(__name__)


class NaiveTextSplitterCalculator(TextSplitterLengthCalculatorProtocol):
    """
    Text splitter length function that estimates how many tokens there
    are in the text rather than actually giving the correct reading.

    The goal is that we have a safe function that can be used accross
    all the popular models.


    Using OpenAI as a guide:
    - https://help.openai.com/en/articles/4936856-what-are-tokens-and-how-to-count-them

    - 4 characters is the average length of one token.
    - 3/4 words is the average length of one token.
    """

    characters_per_token: int | float | decimal.Decimal = 4
    words_per_token: int | float | decimal.Decimal = 0.75
    final_multiplier: int | float | decimal.Decimal = 1

    def get_splitter_length(self, text: str) -> int:
        """
        Estimate how many tokens there are in a piece of text.

        """
        word_count = len(re.findall(r"[^\w\s]|\w+", text))
        char_count = len(text)

        # Estimating 1 token per 4 characters.
        token_char_count = math.ceil(char_count / self.characters_per_token)

        # Estimating 1 token per 3/4 words.
        token_word_count = math.ceil(word_count * self.words_per_token)

        logger.debug(
            "Text splitting: token_char_count=%d token_word_count=%d char_count=%d word_count=%d",
            token_char_count,
            token_word_count,
            char_count,
            word_count,
        )

        return math.ceil(
            max(token_char_count, token_word_count) * self.final_multiplier
        )



import uuid

from django.db import models
from django.utils.translation import gettext_lazy as _
from wagtail.search import index

from wagtail_ai.prompts import DEFAULT_PROMPTS


class Prompt(models.Model, index.Indexed):
    class Method(models.TextChoices):
        REPLACE = "replace", _("Replace content")
        APPEND = "append", _("Append after existing content")

    uuid = models.UUIDField(default=uuid.uuid4, unique=True, editable=False)
    default_prompt_id = models.SmallIntegerField(unique=True, editable=False, null=True)
    label = models.CharField(max_length=50)
    description = models.CharField(
        max_length=255,
        blank=True,
        help_text=_(
            "The prompt description, displayed alongside its label on the settings listing."
        ),
    )
    prompt = models.TextField(  # noqa: DJ001
        null=True,  # field is nullable to manage default prompts, see the prompt_value property for more info.
        blank=False,
        help_text=_(
            "The prompt text sent to the Large Language Model (e.g. ChatGPT) to generate content."
        ),
    )
    method = models.CharField(
        max_length=25,
        choices=Method.choices,
        help_text=_("The method used for processing the responses to the prompt."),
    )

    search_fields = [
        index.AutocompleteField("label"),
        index.SearchField("description"),
        index.SearchField("prompt"),
    ]

    def __str__(self):
        return self.label

    def get_default_prompt_value(self) -> str:
        """
        Return the default prompt value from DEFAULT_PROMPTS.
        """
        return next(
            (
                prompt["prompt"]
                for prompt in DEFAULT_PROMPTS
                if prompt["default_prompt_id"] == self.default_prompt_id
            ),
            "",
        )

    @property
    def is_default(self) -> bool:
        """
        Returns True if the prompt is one of the default prompts.
        """
        return self.default_prompt_id is not None

    @property
    def prompt_value(self) -> str:
        """
        Return the prompt value, otherwise if the prompt is None and belongs
        to the default prompts, map to the default prompt value.
        """
        if self.prompt is None:
            if self.is_default:
                return self.get_default_prompt_value()
            else:
                raise ValueError("Prompt value is None and not a default prompt.")
        return self.prompt



"""
Source:
https://github.com/langchain-ai/langchain/blob/b4312aac5c0567088353178fb70fdb356b372e12/libs/langchain/langchain/text_splitter.py

This code has been copied from the Langchain project and modified.
This code is licensed under the following terms:

MIT License

Copyright (c) LangChain, Inc.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import logging
import re
from collections.abc import Callable, Iterable, MutableSequence, Sequence

from ..types import TextSplitterProtocol

logger = logging.getLogger(__name__)


def _split_text_with_regex(
    text: str, separator: str, keep_separator: bool
) -> list[str]:
    # Now that we have the separator, split the text
    if separator:
        if keep_separator:
            # The parentheses in the pattern keep the delimiters in the result.
            _splits = re.split(f"({separator})", text)
            splits = [_splits[i] + _splits[i + 1] for i in range(1, len(_splits), 2)]
            if len(_splits) % 2 == 0:
                splits += _splits[-1:]
            splits = [_splits[0], *splits]
        else:
            splits = re.split(separator, text)
    else:
        splits = list(text)
    return [s for s in splits if s != ""]


class LangchainRecursiveCharacterTextSplitter(TextSplitterProtocol):
    """Splitting text by recursively look at characters.

    Recursively tries to split by different characters to find one
    that works.

    This class comes from the Langchain project and has been slightly modified.
    """

    separators: Sequence[str]
    chunk_size: int
    chunk_overlap: int = 200
    length_function: Callable[[str], int]
    keep_separator = False
    strip_whitespace: bool = True

    def __init__(
        self,
        *,
        chunk_size: int,
        length_function: Callable[[str], int],
    ) -> None:
        self.separators = ["\n\n", "\n", " ", ""]
        self.length_function = length_function
        self.chunk_size = chunk_size

    def split_text(self, text: str) -> list[str]:
        return self._split_text(text, self.separators)

    def _split_text(self, text: str, separators: Sequence[str]) -> list[str]:
        """Split incoming text and return chunks."""
        final_chunks = []
        # Get appropriate separator to use
        separator = separators[-1]
        new_separators = []
        for i, _s in enumerate(separators):
            _separator = re.escape(_s)
            if _s == "":
                separator = _s
                break
            if re.search(_separator, text):
                separator = _s
                new_separators = separators[i + 1 :]
                break

        splits = _split_text_with_regex(text, separator, self.keep_separator)

        # Now go merging things, recursively splitting longer texts.
        _good_splits = []
        for s in splits:
            if self.length_function(s) < self.chunk_size:
                _good_splits.append(s)
            else:
                if _good_splits:
                    merged_text = self._merge_splits(_good_splits, separator)
                    final_chunks.extend(merged_text)
                    _good_splits = []
                if not new_separators:
                    final_chunks.append(s)
                else:
                    other_info = self._split_text(s, new_separators)
                    final_chunks.extend(other_info)
        if _good_splits:
            merged_text = self._merge_splits(_good_splits, separator)
            final_chunks.extend(merged_text)
        return final_chunks

    def _merge_splits(self, splits: Iterable[str], separator: str) -> list[str]:
        # We now want to combine these smaller pieces into medium size
        # chunks to send to the LLM.
        separator_len = self.length_function(separator)

        docs = []
        current_doc: MutableSequence[str] = []
        total = 0
        for d in splits:
            _len = self.length_function(d)
            if (
                total + _len + (separator_len if len(current_doc) > 0 else 0)
                > self.chunk_size
            ):
                if total > self.chunk_size:
                    logger.warning(
                        f"Created a chunk of size {total}, "
                        f"which is longer than the specified {self.chunk_size}"
                    )
                if len(current_doc) > 0:
                    doc = self._join_docs(current_doc, separator)
                    if doc is not None:
                        docs.append(doc)
                    # Keep on popping if:
                    # - we have a larger chunk than in the chunk overlap
                    # - or if we still have any chunks and the length is long
                    while total > self.chunk_overlap or (
                        total + _len + (separator_len if len(current_doc) > 0 else 0)
                        > self.chunk_size
                        and total > 0
                    ):
                        total -= self.length_function(current_doc[0]) + (
                            separator_len if len(current_doc) > 1 else 0
                        )
                        current_doc = current_doc[1:]
            current_doc.append(d)
            total += _len + (separator_len if len(current_doc) > 1 else 0)
        doc = self._join_docs(current_doc, separator)
        if doc is not None:
            docs.append(doc)
        return docs

    def _join_docs(self, docs: Sequence[str], separator: str) -> str | None:
        text = separator.join(docs)
        if self.strip_whitespace:
            text = text.strip()
        if text == "":
            return None
        else:
            return text



import os
from collections.abc import Mapping
from typing import Any, NotRequired, Self

import llm
from llm.models import dataclass

from ..types import AIResponse
from .base import AIBackend, BaseAIBackendConfig, BaseAIBackendConfigSettings


class LLMBackendConfigSettingsDict(BaseAIBackendConfigSettings):
    PROMPT_KWARGS: NotRequired[Mapping[str, Any] | None]
    INIT_KWARGS: NotRequired[Mapping[str, Any] | None]


@dataclass(kw_only=True)
class LLMBackendConfig(BaseAIBackendConfig[LLMBackendConfigSettingsDict]):
    prompt_kwargs: Mapping[str, Any]
    init_kwargs: Mapping[str, Any]

    @classmethod
    def from_settings(cls, config: LLMBackendConfigSettingsDict, **kwargs: Any) -> Self:
        init_kwargs = config.get("INIT_KWARGS")
        if init_kwargs is None:
            init_kwargs = {}
        kwargs.setdefault("init_kwargs", init_kwargs)

        prompt_kwargs = config.get("PROMPT_KWARGS")
        if prompt_kwargs is None:
            prompt_kwargs = {}
        kwargs.setdefault("prompt_kwargs", prompt_kwargs)

        return super().from_settings(config, **kwargs)


class LLMBackend(AIBackend[LLMBackendConfig]):
    config_cls = LLMBackendConfig

    def prompt_with_context(
        self, *, pre_prompt: str, context: str, post_prompt: str | None = None
    ) -> AIResponse:
        model = self.get_llm_model()
        parts = [pre_prompt, context]

        if post_prompt is not None:
            parts.append(post_prompt)

        full_prompt = os.linesep.join(parts)
        prompt_kwargs = {}
        if self.config.prompt_kwargs is not None:
            prompt_kwargs.update(self.config.prompt_kwargs)
        return model.prompt(full_prompt, **prompt_kwargs)

    def get_llm_model(self) -> llm.Model:
        model = llm.get_model(self.config.model_id)
        if self.config.init_kwargs is not None:
            for config_key, config_val in self.config.init_kwargs.items():
                setattr(model, config_key, config_val)
        return model



# Generated by Django 4.2.7 on 2023-12-08 09:39

import uuid

import wagtail.search.index
from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="Prompt",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "uuid",
                    models.UUIDField(default=uuid.uuid4, editable=False, unique=True),
                ),
                (
                    "default_prompt_id",
                    models.SmallIntegerField(editable=False, null=True, unique=True),
                ),
                ("label", models.CharField(max_length=50)),
                (
                    "description",
                    models.CharField(
                        blank=True,
                        help_text="The prompt description, displayed alongside its label on the settings listing.",
                        max_length=255,
                    ),
                ),
                (
                    "prompt",
                    models.TextField(
                        help_text="The prompt text sent to the Large Language Model (e.g. ChatGPT) to generate content.",
                        null=True,
                    ),
                ),
                (
                    "method",
                    models.CharField(
                        choices=[
                            ("replace", "Replace content"),
                            ("append", "Append after existing content"),
                        ],
                        help_text="The method used for processing the responses to the prompt.",
                        max_length=25,
                    ),
                ),
            ],
            bases=(models.Model, wagtail.search.index.Indexed),
        ),
    ]



import logging
import os
from typing import Type, cast

from django import forms
from django.conf import settings
from django.http import JsonResponse
from django.shortcuts import get_object_or_404
from django.utils.translation import gettext as _
from django.views.decorators.csrf import csrf_exempt
from wagtail.admin.ui.tables import UpdatedAtColumn
from wagtail.admin.viewsets.model import ModelViewSet
from wagtail.images.models import AbstractImage
from wagtail.images.permissions import get_image_model

from . import ai, types
from .ai.base import BackendFeature
from .forms import DescribeImageApiForm, PromptForm
from .models import Prompt

logger = logging.getLogger(__name__)


class AIHandlerException(Exception):
    pass


def _process_backend_request(
    ai_backend: ai.AIBackend, pre_prompt: str, context: str
) -> types.AIResponse:
    """
    Method for processing prompt requests and handling errors.

    Errors will either be an API or Python library error, this method uses exception
    chaining to retain the original error and raise a more generic error message to be sent to the front-end.

    :return: The response message from the AI backend.
    :raises AIHandlerException: Raised for specific error scenarios to be communicated to the front-end.
    """
    try:
        response = ai_backend.prompt_with_context(
            pre_prompt=pre_prompt, context=context
        )
    except Exception as e:
        # Raise a more generic error to send to the front-end
        raise AIHandlerException(
            "Error processing request, Please try again later."
        ) from e
    return response


def _replace_handler(*, prompt: Prompt, text: str) -> str:
    ai_backend = ai.get_backend()
    splitter = ai_backend.get_text_splitter()
    texts = splitter.split_text(text)

    for split in texts:
        response = _process_backend_request(
            ai_backend, pre_prompt=prompt.prompt_value, context=split
        )
        # Remove extra blank lines returned by the API
        message = os.linesep.join([s for s in response.text().splitlines() if s])
        text = text.replace(split, message)

    return text


def _append_handler(*, prompt: Prompt, text: str) -> str:
    ai_backend = ai.get_backend()
    length_calculator = ai_backend.get_splitter_length_calculator()
    if length_calculator.get_splitter_length(text) > ai_backend.config.token_limit:
        raise AIHandlerException("Cannot run completion on text this long")

    response = _process_backend_request(
        ai_backend, pre_prompt=prompt.prompt_value, context=text
    )
    # Remove extra blank lines returned by the API
    message = os.linesep.join([s for s in response.text().splitlines() if s])

    return message


def ErrorJsonResponse(error_message, status=500):
    return JsonResponse({"error": error_message}, status=status)


@csrf_exempt
def text_completion(request) -> JsonResponse:
    prompt_form = PromptForm(request.POST)

    if not prompt_form.is_valid():
        return ErrorJsonResponse(prompt_form.errors_for_json_response(), status=400)

    try:
        prompt = Prompt.objects.get(uuid=prompt_form.cleaned_data["prompt"])
    except Prompt.DoesNotExist:
        return ErrorJsonResponse(_("Invalid prompt provided."), status=400)

    handlers = {
        Prompt.Method.REPLACE: _replace_handler,
        Prompt.Method.APPEND: _append_handler,
    }

    handler = handlers[Prompt.Method(prompt.method)]

    try:
        response = handler(prompt=prompt, text=prompt_form.cleaned_data["text"])
    except AIHandlerException as e:
        return ErrorJsonResponse(str(e), status=400)
    except Exception:
        logger.exception("An unexpected error occurred.")
        return ErrorJsonResponse(_("An unexpected error occurred."))

    return JsonResponse({"message": response})


def user_has_permission_for_image(user, image):
    from wagtail.images.permissions import permission_policy

    return permission_policy.user_has_permission_for_instance(user, "choose", image)


def describe_image(request) -> JsonResponse:
    form = DescribeImageApiForm(request.POST)
    if not form.is_valid():
        return ErrorJsonResponse(form.errors_for_json_response(), status=400)

    model = cast(Type[AbstractImage], get_image_model())
    image = get_object_or_404(model, pk=form.cleaned_data["image_id"])

    if not user_has_permission_for_image(request.user, image):
        return ErrorJsonResponse("Access denied", status=403)

    try:
        backend = ai.get_backend(BackendFeature.IMAGE_DESCRIPTION)
    except ai.BackendNotFound:
        return ErrorJsonResponse(
            "No backend is configured for image description. Please set"
            " `IMAGE_DESCRIPTION_BACKEND` in `settings.WAGTAIL_AI`.",
            status=400,
        )

    wagtail_ai_settings = getattr(settings, "WAGTAIL_AI", {})
    rendition_filter = wagtail_ai_settings.get(
        "IMAGE_DESCRIPTION_RENDITION_FILTER", "max-800x600"
    )
    rendition = image.get_rendition(rendition_filter)

    maxlength = form.cleaned_data["maxlength"]
    prompt = wagtail_ai_settings.get("IMAGE_DESCRIPTION_PROMPT")

    if prompt is None:
        prompt = (
            "Describe this image. Make the description suitable for use as an alt-text."
        )
        if maxlength is not None:
            prompt += f" Make the description less than {maxlength} characters long."

    try:
        ai_response = backend.describe_image(image_file=rendition.file, prompt=prompt)
        description = ai_response.text()
    except Exception:
        logger.exception("There was an issue describing the image.")
        return ErrorJsonResponse("There was an issue describing the image.")

    if not description:
        return ErrorJsonResponse("There was an issue describing the image.")

    if maxlength is not None:
        description = description[:maxlength]

    return JsonResponse({"message": description})


class PromptEditForm(forms.ModelForm):
    """
    Custom form for the model admin to allow users to view and edit default prompts
    """

    class Meta:
        model = Prompt
        fields = ["label", "description", "prompt", "method"]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if self.instance.is_default:
            # Make the prompt field not required if it's a default prompt
            self.fields["prompt"].required = False
            # Populate the placeholder with the value from DEFAULT_PROMPTS
            self.fields["prompt"].widget.attrs[
                "placeholder"
            ] = self.instance.get_default_prompt_value()


class PromptViewSet(ModelViewSet):
    model = Prompt
    form_fields = PromptEditForm.Meta.fields
    list_display = ["label", "description", "method", UpdatedAtColumn()]  # type: ignore
    icon = "edit"
    add_to_settings_menu = True
    menu_order = 300

    def get_form_class(self, for_update=False):
        if for_update:
            return PromptEditForm
        return super().get_form_class(for_update)


prompt_viewset = PromptViewSet("prompt")



from collections.abc import Callable, Iterator
from typing import Any, Protocol


class AIResponse(Protocol):
    """
    Compatible with the llm.Response.
    """

    def __iter__(self) -> Iterator[str]:
        ...

    def text(self) -> str:
        ...


class TextSplitterProtocol(Protocol):
    def __init__(
        self, *, chunk_size: int, length_function: Callable[[str], int], **kwargs: Any
    ) -> None:
        ...

    def split_text(self, text: str) -> list[str]:
        ...


class TextSplitterLengthCalculatorProtocol(Protocol):
    def get_splitter_length(self, text: str) -> int:
        ...



# Created using Django 4.2.7 on 2023-12-08 09:39

from django.db import migrations

from wagtail_ai.prompts import DEFAULT_PROMPTS


def set_default_ai_prompts(apps, schema_editor):
    """
    Loop through each default prompt and
    populate the database.
    """
    Prompt = apps.get_model("wagtail_ai", "Prompt")
    for default_prompt in DEFAULT_PROMPTS:
        Prompt.objects.update_or_create(
            default_prompt_id=default_prompt["default_prompt_id"],
            # The prompt field is left blank to allow users to override default prompt values and maintainers
            # to manage default prompts that are None (not overridden) using the get_default_prompt_value method.
            prompt=None,
            defaults={
                "label": default_prompt["label"],
                "description": default_prompt.get("description", ""),
                "method": default_prompt.get("method", None),
            },
        )


class Migration(migrations.Migration):
    dependencies = [
        ("wagtail_ai", "0001_initial"),
    ]

    operations = [
        migrations.RunPython(set_default_ai_prompts, migrations.RunPython.noop),
    ]
